#include "App.h"

#include "AppState.h"
#include "MissionSelect.h"
#include "events.h"

#include <SDL2/SDL.h>
#include <SDL2/SDL_events.h>
#include <SDL2/SDL_image.h>
#include <SDL2/SDL_keycode.h>
#include <SDL2/SDL_render.h>
#include <SDL2/SDL_timer.h>

#include <fmt/format.h>
#include <spdlog/spdlog.h>

#include <memory>
#include <queue>
#include <stdexcept>

App::App() {
  init_sdl();
  init_state();
}

App::~App() {
  // Destroy window
  SDL_DestroyRenderer(renderer);
  renderer = nullptr;
  SDL_DestroyWindow(window);
  window = nullptr;
}

void App::init_sdl() {
  window = SDL_CreateWindow("FT22 STW Display", SDL_WINDOWPOS_UNDEFINED,
                            SDL_WINDOWPOS_UNDEFINED, SCREEN_WIDTH,
                            SCREEN_HEIGHT, SDL_WINDOW_SHOWN);
  if (window == nullptr) {
    throw std::runtime_error(
        fmt::format("Couldn't create window: {}", SDL_GetError()));
  }
  // Software renderer is *MUCH* (~50x speedup) faster than hardware accelerated
  // renderer on a Pi Zero.
  renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_SOFTWARE);
  if (renderer == nullptr) {
    throw std::runtime_error(
        fmt::format("Couldn't create renderer: {}", SDL_GetError()));
  }
}

void App::init_state() {
  try {
    state = std::make_unique<AppState>("/dev/i2c-3");
  } catch (const std::runtime_error& e) {
    spdlog::error("Couldn't create AppState: {}", e.what());
  }
}

int App::run() {
  mission_select = std::make_unique<MissionSelect>(renderer);

  running = true;

  uint32_t second_start = SDL_GetTicks();
  unsigned frames = 0;
  while (running) {
    uint32_t now = SDL_GetTicks();
    if (now - second_start > 5000) {
      spdlog::info("{} FPS", frames / 5);
      second_start = now;
      frames = 0;
    }

    if (state != nullptr) {
      state->poll();
    }

    handle_events();
    render();

    frames++;
    SDL_Delay(1);
  }

  return 0;
}

void App::handle_events() {
  std::queue<Event> events;
  SDL_Event e;
  while (SDL_PollEvent(&e) != 0) {
    if (e.type == SDL_QUIT) {
      running = false;
    } else if (e.type == SDL_KEYUP) {
      switch (e.key.keysym.sym) {
      case SDLK_DOWN:
      case SDLK_RIGHT:
        events.push(Event::Next);
        break;
      case SDLK_UP:
      case SDLK_LEFT:
        events.push(Event::Prev);
        break;
      case SDLK_RETURN:
      case SDLK_KP_ENTER:
        events.push(Event::Confirm);
        break;
        // We can just ignore other keypresses, so no need for a default clause
      }
    }
  }

  AppView view = state->get_view();
  switch (view) {
  case AppView::MISSION_SELECT:
    mission_select->handle_events(events);
    break;
  default:
    throw std::runtime_error(fmt::format("Unknown view: {}", view));
  }
}

void App::render() {
  AppView view = state->get_view();
  switch (view) {
  case AppView::MISSION_SELECT:
    mission_select->draw();
    break;
  default:
    throw std::runtime_error(fmt::format("Unknown view: {}", view));
  }
  SDL_RenderPresent(renderer);
}

SDLManager::SDLManager() {
  if (SDL_Init(SDL_INIT_VIDEO) < 0) {
    throw std::runtime_error(
        fmt::format("Couldn't initialize SDL: {}", SDL_GetError()));
  }
  if (TTF_Init() == -1) {
    throw std::runtime_error(
        fmt::format("Couldn't initialize SDL_ttf: {}", TTF_GetError()));
  }
  int img_flags = IMG_INIT_PNG;
  if (!(IMG_Init(img_flags) & img_flags)) {
    throw std::runtime_error(
        fmt::format("Couldn't initialitze SDL_image: {}", IMG_GetError()));
  }
}

SDLManager::~SDLManager() {
  IMG_Quit();
  TTF_Quit();
  SDL_Quit();
}